/* Peer - All public information concerning a peer.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.io.*;
import java.net.*;
import java.util.Arrays;
import java.util.Map;

import org.klomp.snark.bencode.*;

public class Peer
{
  // Identifying property, the peer id of the other side.
  private final byte[] id;

  private final String ip;
  private final int port;

  private final byte[] my_id;
  private final MetaInfo metainfo;

  // The data in/output streams set during the handshake and used by
  // the actual connections.
  private DataInputStream din;
  private DataOutputStream dout;

  // Keeps state for in/out connections.  Non-null when the handshake
  // was successful, the connection setup and runs
  PeerState state;

  boolean downloading;
  boolean uploading;

  private boolean deregister = true;

  /**
   * Creates a disconnected peer from a bencoded dictionary.
   */
  public Peer(InputStream in, byte[] my_id, MetaInfo metainfo)
    throws IOException
  {
    this(new BDecoder(in), my_id, metainfo);
  }

  /**
   * Creates a disconnected peer from a bencoded dictionary.
   */
  public Peer(BDecoder be, byte[] my_id, MetaInfo metainfo)
    throws IOException
  {
    this(be.bdecodeMap().getMap(), my_id, metainfo);
  }

  /**
   * Creates a disconnected peer from a Map containing peer id, ip and
   * port.
   */
  public Peer(Map m, byte[] my_id, MetaInfo metainfo)
    throws InvalidBEncodingException
  {
    BEValue bevalue = (BEValue)m.get("peer id");
    if (bevalue == null)
      throw new InvalidBEncodingException("peer id missing");
    id = bevalue.getBytes();

    bevalue = (BEValue)m.get("ip");
    if (bevalue == null)
      throw new InvalidBEncodingException("ip missing");
    ip = bevalue.getString();

    bevalue = (BEValue)m.get("port");
    if (bevalue == null)
      throw new InvalidBEncodingException("port missing");
    port = bevalue.getInt();

    this.my_id = my_id;
    this.metainfo = metainfo;
  }

  /**
   * Creates a unconnected peer from the input and output stream got
   * from the socket. Note that the complete handshake (which can take
   * some time or block indefinitely) is done in the calling Thread to
   * get the remote peer id. To completely start the connection call
   * the connect() method.
   *
   * @exception IOException when an error occurred during the handshake.
   */
  public Peer(final Socket sock, byte[] my_id, MetaInfo metainfo)
    throws IOException
  {
    this.ip = sock.getInetAddress().toString();
    this.port = sock.getPort();

    this.my_id = my_id;
    this.metainfo = metainfo;

    this.id = handshake(sock);
  }

  /**
   * Returns the id of the peer.
   */
  public byte[] getID()
  {
    return id;
  }

  public String getIP()
  {
    return ip;
  }

  public int getPort()
  {
    return port;
  }

  public String toString()
  {
    return "Peer[" + (id != null ? idencode(id) : "null")
      + " " + ip + ":" + port + "]";
  }

  /**
   * Encode an id as a hex encoded string and remove leading zeros.
   */
  private static String idencode(byte[] bs)
  {
    boolean leading_zeros = true;

    StringBuffer sb = new StringBuffer(bs.length*2);
    for (int i = 0; i < bs.length; i++)
      {
	int c = bs[i] & 0xFF;
	if (leading_zeros && c == 0)
	  continue;
	else
	  leading_zeros = false;

	if (c < 16)
	  sb.append('0');
	sb.append(Integer.toHexString(c));
      }

    return sb.toString();
  }

  /**
   * The hash code of a Peer is the exclusive or of all id bytes.
   */
  public int hashCode()
  {
    int b = 0;
    for (int i = 0; i < id.length; i++)
      b ^= id[i];
    return b;
  }

  /**
   * Two Peers are equal when they have the same id.
   * All other properties are ignored.
   */
  public boolean equals(Object o)
  {
    if (o instanceof Peer)
      {
	Peer p = (Peer)o;

	boolean equal = true;
	for (int i = 0; equal && i < id.length; i++)
	  equal = id[i] == p.id[i];
	return equal;
      }
    else
      return false;
  }

  /**
   * Runs the connection to the other peer. This method does not
   * return until the connection is terminated.
   *
   * When the connection is correctly started the connected() method
   * of the given PeerListener is called. If the connection ends or
   * the connection could not be setup correctly the disconnected()
   * method is called.
   *
   * If the given BitField is non-null it is send to the peer as first
   * message.
   */
  public void runConnection(PeerListener listener, BitField bitfield)
  {
    if (state != null)
      throw new IllegalStateException("Peer already started");

    try
      {
	// Do we need to handshake?
	if (din == null)
	  {
	    Socket sock = new Socket(ip, port);
	    byte [] peerID = handshake(sock);
	    if (!Arrays.equals(peerID, id))
	      throw new IOException("Unexpected peerID '"
				    + idencode(peerID)
				    + "' expected '"
				    + idencode(id) + "'");
	  }
	
	PeerConnectionIn in = new PeerConnectionIn(this, din);
	PeerConnectionOut out = new PeerConnectionOut(this, dout);
	PeerState s = new PeerState(this, listener, metainfo, in, out);
	
	// Send our bitmap
	if (bitfield != null)
	  s.out.sendBitfield(bitfield);
    
	// We are up and running!
	state = s;
	listener.connected(this);
  
	// Use this thread for running the incomming connection.
	// The outgoing connection has created its own Thread.
	s.in.run();
      }
    catch(IOException eofe)
      {
	// Ignore, probably just the other side closing the connection.
	// Or refusing the connection, timing out, etc.
      }
    catch(Throwable t)
      {
	System.err.println(this + ": " + t);
	t.printStackTrace();
      }
    finally
      {
	// XXX - Actually Wait till it is all over...?
	if (deregister) listener.disconnected(this);
      }
  }

  /**
   * Sets DataIn/OutputStreams, does the handshake and returns the id
   * reported by the other side.
   */
  private byte[] handshake(Socket sock) throws IOException
  {
    din = new DataInputStream
      (new BufferedInputStream(sock.getInputStream()));
    dout = new DataOutputStream
      (new BufferedOutputStream(sock.getOutputStream()));
    
    // Handshake write - header
    dout.write(19);
    dout.write("BitTorrent protocol".getBytes("UTF-8"));
    // Handshake write - zeros
    byte[] zeros = new byte[8];
    dout.write(zeros);
    // Handshake write - metainfo hash
    byte[] shared_hash = metainfo.getInfoHash();
    dout.write(shared_hash);
    // Handshake write - peer id
    dout.write(my_id);
    dout.flush();
    
    // Handshake read - header
    byte b = din.readByte();
    if (b != 19)
      throw new IOException("Handshake failure, expected 19, got "
			    + (b & 0xff));
    
    byte[] bs = new byte[19];
    din.readFully(bs);
    String bittorrentProtocol = new String(bs, "UTF-8");
    if (!"BitTorrent protocol".equals(bittorrentProtocol))
      throw new IOException("Handshake failure, expected "
			    + "'Bittorrent protocol', got '"
			    + bittorrentProtocol + "'");
    
    // Handshake read - zeros
    din.readFully(zeros);
    
    // Handshake read - metainfo hash
    bs = new byte[20];
    din.readFully(bs);
    if (!Arrays.equals(shared_hash, bs))
      throw new IOException("Unexpected MetaInfo hash");

    // Handshake read - peer id
    din.readFully(bs);
    return bs;
  }

  public boolean isConnected()
  {
    return state != null;
  }

  /**
   * Disconnects this peer if it was connected.  If deregister is
   * true, PeerListener.disconnected() will be called when the
   * connection is completely terminated. Otherwise the connection is
   * silently terminated.
   */
  public void disconnect(boolean deregister)
  {
    // Both in and out connection will call this.
    this.deregister = deregister;
    disconnect();
  }

  void disconnect()
  {
    PeerState s = state;
    if (s != null)
      {
	PeerConnectionIn in = s.in;
	if (in != null)
	  in.disconnect();
	PeerConnectionOut out = s.out;
	if (out != null)
	  out.disconnect();
      }
    state = null;
  }

  /**
   * Wheter or not to (start) uploading to this peer.
   */
  public void upload(boolean upload)
  {
    uploading = upload;
    PeerState s = state;
    if (s != null)
      s.upload(upload);
  }

  public boolean isUpLoading()
  {
    return uploading;
  }

  /**
   * Wheter or not to (start) downloading from this peer.
   */
  public void download(boolean download)
  {
    downloading = download;
    PeerState s = state;
    if (s != null)
      s.download(download);
  }

  public boolean isDownLoading()
  {
    return downloading;
  }

  /**
   * Tell the peer we have another piece.
   */
  public void have(int piece)
  {
    PeerState s = state;
    if (s != null)
      {
	PeerConnectionOut out = s.out;
	if (out != null)
	  out.sendHave(piece);
      }
  }

  /**
   * Whether or not the peer is interested in pieces we have.
   */
  public boolean isInterested()
  {
    PeerState s = state;
    return (s != null) && s.interested;
  }

  /**
   * Wheter or not the peer has pieces we want from it.
   */
  public boolean isInteresting()
  {
    PeerState s = state;
    return (s != null) && s.interesting;
  }

  /**
   * Wheter or not we are choking the peer.
   */
  public boolean isChoking()
  {
    PeerState s = state;
    return (s != null) && s.choking;
  }

  /**
   * Wheter or not the peer choked us.
   */
  public boolean isChoked()
  {
    PeerState s = state;
    return (s != null) && s.choked;
  }

  /**
   * Returns the number of bytes that have been downloaded.
   * Can be reset to zero with <code>resetCounters()</code>/
   */
  public long getDownloaded()
  {
    PeerState s = state;
    return (s != null) ? s.downloaded : 0;
  }

  /**
   * Returns the number of bytes that have been uploaded.
   * Can be reset to zero with <code>resetCounters()</code>/
   */
  public long getUploaded()
  {
    PeerState s = state;
    return (s != null) ? s.uploaded : 0;
  }

  /**
   * Resets the downloaded and uploaded counters to zero.
   */
  public void resetCounters()
  {
    PeerState s = state;
    if (s != null)
      {
	s.downloaded = 0;
	s.uploaded = 0;
      }
  }
}
